/* Based on asbutton which was based on wmbutton which was based on other
 * things
 */

/* TODO:	enable icon switching. See State().
 * 		create/find proper icons.
 *		allow normal drag/right clicking.
 */

#include <Xlib.h>
#include <Xutil.h>
#include <xpm.h>
#include <extensions/shape.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>

#include "asdial.h"

#include "connecting.xpm"
#include "connected.xpm"
#include "failed.xpm"
#include "disconnected.xpm"

static volatile sig_atomic_t sighandler_write_pipe;

int main(int argc, char **argv) 
{
	char *command;
	char *geometry;
	char *display_str;
	pid_t dialpid = 0;
	int readp;
	int alive = 1;

	if (defaultargs(&command, &geometry, &display_str))
		return 1;
	if (getargs(argc, argv, &command, &geometry, &display_str))
		return 1;
	/* X stuff */
	display = XOpenDisplay(display_str);
	if (display==NULL) {
		fprintf(stderr,"Fail: XOpenDisplay for '%s'\n", display_str);	
		return 1;
	}
	XSetup(geometry);
	XSetCommand(display, win, argv, argc);	
	XMapWindow(display, win);

	readp = sigsetup();
	State(STATE_DISCONNECTED);
	while(alive) {
		while (XPending(display)) {
			if (HandleEvent(&dialpid, command))
				alive = 0;
		}
		switch(getsignal(readp)) {
		case SIG_ERROR:
			alive = 0;
		case SIG_TIMEOUT:
			break;
		case SIG_CHILDDEAD:
			if (dialpid) {
				dialpid = 0;
				State(STATE_FAILED);
			}
	       		break;
		case SIG_CONNECTED:
			State(STATE_CONNECTED);
			break;
		}
	}
	XFreeGC(display, gc);
	XDestroyWindow(display,win);
	XDestroyWindow(display,iconwin);
	XCloseDisplay(display);
	if (dialpid)
		kill(0, SIGTERM);
	return 0;
}

int getsignal(int readp)
{
	int ret;
	char c;
	fd_set readers;
	struct timeval tv;

	tv.tv_sec = 0;
	tv.tv_usec = 1000; /* TODO: check */

	FD_ZERO(&readers);
	FD_SET(readp, &readers);

	do {
		ret = select(readp+1, &readers, NULL, NULL, &tv);
	} while (ret==-1 && errno==EINTR);
	switch(ret) {
	case -1:
		perror("select failed"); 
		return SIG_ERROR;
	case  0:
		return SIG_TIMEOUT;
	}
	do {
		ret = read(readp, &c, 1);
	} while (ret==-1 && errno==EINTR);
	if (ret==-1) {
		perror("read failed");
		return SIG_ERROR;
	}
	if (c=='d')
		return SIG_CHILDDEAD;
	return SIG_CONNECTED;
}

int sigsetup(void)
{
	int fd[2];
	struct sigaction cursig;
	if (pipe(fd)) {
		perror("pipe failed");
		return -1;
	}
	sighandler_write_pipe = fd[1];

	memset(&cursig, 0, sizeof(cursig));
	cursig.sa_handler = sighandler;
	if (-1==sigaction(SIGCHLD, &cursig, NULL) ||
			-1==sigaction(SIGUSR1, &cursig, NULL) ||
			-1==sigaction(SIGTERM, &cursig, NULL)) {
		fprintf(stderr, "sigaction failed\n");
		close(fd[0]); close(fd[1]);
		return -1;
	}
	return fd[0];
}

void sighandler(int signo)
{
	char c;
	if (signo==SIGCHLD) {
		c = 'd';
		wait(NULL);
		write(sighandler_write_pipe, &c, 1);
	} else if (signo==SIGUSR1) {
		c = 'c';
		write(sighandler_write_pipe, &c, 1);
	} else if (signo==SIGTERM) {
		kill(0, SIGTERM);
		_exit(0);
	}
	return;
}

int HandleEvent(pid_t *dialpid, const char *command)
{
	XEvent report;

	XNextEvent(display, &report);
	switch (report.type) {
	case Expose:
		if (report.xexpose.count==0) 	
			Redraw();
		break;      
	case ConfigureNotify:
		Redraw();
		break;
	case ButtonPress:
		if (report.xbutton.button!=Button1)
			break;
		if (*dialpid) {
			kill(*dialpid, SIGTERM);
			waitpid(*dialpid, NULL, 0);
			*dialpid = 0;
			State(STATE_DISCONNECTED);
			break;
		}
		*dialpid = RunApp(command);
		State(STATE_CONNECTING);
		break;
	case DestroyNotify:
		return -1;
	}
	return 0;
}

void State(int state)
{
	/* TODO:
	 * for STATE_DISCONNECTED change to disconnected_xpm
	 * for STATE_CONNECTED change to connected_xpm
	 * for STATE_CONNECTING change to connecting_xpm
	 * for STATE_FAILED change to failed_xpm
	 */
	Redraw();
	return;
}

pid_t RunApp(const char *command)
{
#define MAXLEN 256
	pid_t pid;
	char ourpid[MAXLEN];

	if (-1==snprintf(ourpid, MAXLEN, "PARENT_PID=%ld", (long)getpid()))
		return -1;
	pid = fork();
	if (pid==0) {
		putenv(ourpid);
		execlp("/bin/sh", "sh", "-c", command, NULL);
	       	_exit(-1);
	}
	return pid;
}
	
int defaultargs(char **command, char **geometry, char **display_str)
{
	*command = *geometry = *display_str = NULL;
	if (getargs_set(command, "dial -c \"kill -USR1 $PARENT_PID\""))
		return 1;
	if (getargs_set(geometry, "64x64+0+0"))
		return 1;
	if (getargs_set(display_str, ""))
		return 1;
	return 0;
}

int getargs(int argc, char **argv, char **command, char **geometry,
		char **display_str)
{
	int ret;
	while ((ret=getopt(argc, argv, "c:g:dh"))!=EOF) {
		switch(ret) {
		case 'c':
			if (getargs_set(command, optarg))
				return 1;
			break;
		case 'g':
			if (getargs_set(geometry, optarg))
				return 1;
			break;
		case 'd':
			if (getargs_set(display_str, optarg))
				return 1;
			break;
		case 'h':
			helpinfo();
			break;
		}
	}
	return 0;
}

int getargs_set(char **to, const char *from)
{
	free(*to);
	*to = strdup(from);
	if (*to==NULL) {
		perror("malloc failed");
		return 1;
	}
	return 0;
}
		
/* TODO */
void helpinfo(void)
{
	fprintf(stderr, "TODO\n");
	return;
}

/***********************************************************************
 *   XSetup            
 *
 *   Initializes X Window stuff 
 *
 ***********************************************************************/

void XSetup(const char *geometry)
{
	int dummy = 0;
	XSizeHints xsizehints;
	XWMHints  *xwmhints;
	XGCValues  xgcValues;	
	XTextProperty app_name_atom;

	screen  = DefaultScreen(display);
	rootwin = RootWindow(display, screen);
	depth   = DefaultDepth(display, screen);

	bg_pixel = WhitePixel(display, screen); 
	fg_pixel = BlackPixel(display, screen); 

	xsizehints.flags  = USSize | USPosition;
	xsizehints.width  = 64;
	xsizehints.height = 64;

	/* Parse Geometry string and fill in sizehints fields */
	XWMGeometry(display, screen, geometry, NULL, 0, 
	      &xsizehints, &xsizehints.x, &xsizehints.y,
	      &xsizehints.width, &xsizehints.height, &dummy);

	CreateWindow(&win, &rootwin, xsizehints);
	CreateWindow(&iconwin, &win, xsizehints); 

	/* Set up shaped windows */
#if 0 /* TODO */
	if ((pixmask = XCreateBitmapFromData(display, 
					 win, mask_bits,
					 mask_width, mask_height)) == 0) {
	  fprintf(stderr,"Fail: XCreateBitmapFromData\n");
	}
#endif
	
	XShapeCombineMask(display, win, ShapeBounding, 0, 0, pixmask, ShapeSet);
	XShapeCombineMask(display, iconwin, ShapeBounding, 0, 0, pixmask, ShapeSet);

	/* Convert in pixmaps from .xpm includes. */
	GetPixmaps();
	
	/* Interclient Communication stuff */
	/* Appicons don't work with out this stuff */
	xwmhints = XAllocWMHints();
	xwmhints->flags = WindowGroupHint | IconWindowHint | StateHint;	
	xwmhints->icon_window = iconwin;
	xwmhints->window_group = win;
	xwmhints->initial_state = WithdrawnState;  
	XSetWMHints(display, win, xwmhints);
	XSetWMNormalHints(display, win, &xsizehints);

	/* Tell window manager what the title bar name is. We never see */
	/* this anyways in the WithdrawnState      */
	if (XStringListToTextProperty(&app_name, 1, &app_name_atom) == 0) {
	  	fprintf(stderr,"%s: Can't set up window name\n", app_name);
	  	exit(-1);
	}
	XSetWMName(display, win, &app_name_atom);
	
	/* Create Graphic Context */	
	if (( gc = XCreateGC(display, win,(GCForeground | GCBackground), &xgcValues))
	     == NULL ) {
	  	fprintf(stderr,"Fail: XCreateGC\n");	
	  	exit(-1);
	}

	/* XEvent Masks. We want both window to process X events */
	XSelectInput(display, win, ExposureMask | PointerMotionMask | ButtonPressMask | StructureNotifyMask );  
	XSelectInput(display, iconwin, ExposureMask | PointerMotionMask | ButtonPressMask | StructureNotifyMask );
}

/***********************************************************************
 *   CreateWindow      
 *
 *   Create a window for XSetup()
 *
 ***********************************************************************/

void CreateWindow(Window *win, Window *rootwin, XSizeHints xsizehints)
{
	XClassHint xclasshint;
	*win = XCreateSimpleWindow(display, *rootwin,
				  xsizehints.x, xsizehints.y,
				  xsizehints.width, xsizehints.height,
				  0, fg_pixel, bg_pixel);
	xclasshint.res_name = "asdial";
	xclasshint.res_class = "ASdial";
	XSetClassHint(display, *win, &xclasshint);
}

/***********************************************************************
 *   Redraw
 *
 *	  Map the button region coordinates.
 *
 *   Draw the appropriate number of buttons on the 'visible' Pixmap 
 *   using data from the 'buttons' pixmap.
 *
 *   Then, copy the 'visible' pixmap to the two windows ( the withdrawn
 *   main window and the icon window which is the main window's icon image.)
 ***********************************************************************/

void Redraw(void)
{
	FlushExpose(iconwin); 
	XCopyArea(display, visible.pixmap, iconwin, gc, 0, 0,
	    visible.attributes.width, visible.attributes.height, 0, 0 );
}

/***********************************************************************
 *   GetPixmaps
 *    
 *   Load XPM data into X Pixmaps.
 *  
 *   Pixmap template contains the untouched window backdrop image.
 *   Pixmap visible is the template pixmap with buttons drawn on it.
 *          -- what is seen by the user.
 *   Pixmap buttons holds the images for individual buttons that are
 *          later copied onto Pixmap visible.
 ***********************************************************************/

void GetPixmaps(void)
{
	//visible.attributes.valuemask  |= (XpmReturnPixels | XpmReturnExtensions);
	/* Visible Pixmap. Copied from template Pixmap and then drawn to. */
#if 0 /* TODO */
	if ( XpmCreatePixmapFromData(	display, rootwin, foo_xpm,
	     &visible.pixmap, &visible.mask, 
	     &visible.attributes) != XpmSuccess ) {
		fprintf(stderr, "Can't Create 'visible' Pixmap");
	   exit(1);
	}
#endif
}

/***********************************************************************
 *   FlushExpose
 ***********************************************************************/

void FlushExpose(Window w) 
{
	XEvent      dummy;

	while (XCheckTypedWindowEvent(display, w, Expose, &dummy));
}
